@@ -1,11 +1,14 @@
 // SPDX-License-Identifier: AGPL-3.0-or-later
-// Copyright (C) 2021  Philipp Emanuel Weidmann <pew@worldwidemann.com>
+// Copyright (C) 2021-2022  Philipp Emanuel Weidmann <pew@worldwidemann.com>
 
 use std::collections::HashMap;
 
 use num::{One, ToPrimitive, Zero};
 
-use crate::expression::{Complex, Expression, RationalRepresentation};
+use crate::{
+    expression::{Complex, Expression, RationalRepresentation},
+    functions::functions,
+};
 
 /// Error that occurred while trying to evaluate an expression.
 #[derive(PartialEq, Eq, Clone, Debug)]
@@ -34,6 +37,27 @@ pub enum Error {
         base: Expression,
         exponent: Expression,
     },
+    /// Vector or matrix expression indexed by an expression that evaluates to
+    /// an integer outside the range of valid indices for that vector or matrix.
+    IndexOutOfBounds {
+        expression: Expression,
+        vector_or_matrix: Expression,
+        index: Expression,
+    },
+    /// Function expression evaluated with a number of arguments
+    /// that is invalid for the function.
+    InvalidNumberOfArguments {
+        expression: Expression,
+        min_number: usize,
+        max_number: usize,
+        given_number: usize,
+    },
+    /// Function expression evaluated with an argument
+    /// that is invalid for the function in that position.
+    InvalidArgument {
+        expression: Expression,
+        argument: Expression,
+    },
 }
 
 impl Expression {
@@ -72,11 +96,14 @@ impl Expression {
             (
                 Variable(_)
                 | Function(_, _)
+                | FunctionValue(_, _)
                 | Integer(_)
                 | Rational(_, _)
                 | Complex(_, _)
                 | Vector(_)
+                | VectorElement(_, _)
                 | Matrix(_)
+                | MatrixElement(_, _, _)
                 | Boolean(_)
                 | Sum(_, _)
                 | Difference(_, _)
@@ -226,16 +253,14 @@ impl Expression {
                                 base: a_original.clone(),
                                 exponent: b_original.clone(),
                             })
+                        } else if let Some(b) = b.to_i32() {
+                            Ok(Complex(a.powi(b), representation))
                         } else {
-                            if let Some(b) = b.to_i32() {
-                                Ok(Complex(a.powi(b), representation))
-                            } else {
-                                // TODO
-                                Ok(Power(
-                                    Box::new(a_evaluated.clone()),
-                                    Box::new(b_evaluated.clone()),
-                                ))
-                            }
+                            // TODO
+                            Ok(Power(
+                                Box::new(a_evaluated.clone()),
+                                Box::new(b_evaluated.clone()),
+                            ))
                         }
                     }
                     Equal(_, _) => Ok(Boolean(a == b)),
@@ -271,6 +296,48 @@ impl Expression {
                 }
             }
 
+            (Sum(_, _) | Difference(_, _), Mat(a), Mat(b)) => {
+                if a.shape() == b.shape() {
+                    Ok(Matrix(match self {
+                        Sum(_, _) => a + b,
+                        Difference(_, _) => a - b,
+                        _ => unreachable!(),
+                    }))
+                } else {
+                    Err(IncompatibleOperands {
+                        expression: self.clone(),
+                        operand_1: a_original.clone(),
+                        operand_2: b_original.clone(),
+                    })
+                }
+            }
+
+            (Product(_, _), Mat(a), Mat(b)) => {
+                if a.is_empty() && b.is_empty() {
+                    Ok(Matrix(a))
+                } else if !a.is_empty() && !b.is_empty() && a.ncols() == b.nrows() {
+                    Ok(Matrix(crate::expression::Matrix::from_fn(
+                        a.nrows(),
+                        b.ncols(),
+                        |i, j| {
+                            (0..a.ncols())
+                                .map(|k| a[(i, k)].clone() * b[(k, j)].clone())
+                                .reduce(|a, b| a + b)
+                                .unwrap()
+                        },
+                    )))
+                } else {
+                    Err(IncompatibleOperands {
+                        expression: self.clone(),
+                        operand_1: a_original.clone(),
+                        operand_2: b_original.clone(),
+                    })
+                }
+            }
+
+            (Product(_, _), Mat(a), _) => Ok(Matrix(a.map(|element| element * b.clone()))),
+            (Product(_, _), _, Mat(b)) => Ok(Matrix(b.map(|element| a.clone() * element))),
+
             (Equal(_, _), Bool(Some(a)), Bool(Some(b))) => Ok(Boolean(a == b)),
             (NotEqual(_, _), Bool(Some(a)), Bool(Some(b))) => Ok(Boolean(a != b)),
             (And(_, _), Bool(Some(a)), Bool(Some(b))) => Ok(Boolean(a && b)),
@@ -294,11 +361,14 @@ impl Expression {
             (
                 Variable(_)
                 | Function(_, _)
+                | FunctionValue(_, _)
                 | Integer(_)
                 | Rational(_, _)
                 | Complex(_, _)
                 | Vector(_)
+                | VectorElement(_, _)
                 | Matrix(_)
+                | MatrixElement(_, _, _)
                 | Boolean(_)
                 | Negation(_)
                 | Not(_),
@@ -313,42 +383,203 @@ impl Expression {
     /// can be used to set the values of variables by their identifiers.
     fn evaluate_step(&self, context: &HashMap<String, Self>) -> Result<Self, Error> {
         use crate::expression::Expression::*;
+        use crate::expression::Type::{
+            Boolean as Bool, Function as Fun, Matrix as Mat, Number as Num,
+        };
+        use Error::*;
 
-        match self {
+        let expression = self.simplify();
+
+        match &expression {
             Variable(identifier) => context
                 .get(identifier)
-                .map_or_else(|| Ok(self.clone()), |x| x.evaluate_step(context)),
-            Function(_, _) => Ok(self.clone()), // TODO
-            Integer(_) => Ok(self.clone()),
+                .map_or_else(|| Ok(expression), |x| x.evaluate_step(context)),
+            Function(_, _) => Ok(expression),
+            FunctionValue(function, arguments) => {
+                let function_original = function;
+
+                let function = function.evaluate_step(context)?;
+
+                let mut arguments_evaluated = Vec::new();
+
+                for argument in arguments {
+                    arguments_evaluated.push(argument.evaluate_step(context)?);
+                }
+
+                match function.typ() {
+                    Num(_, _) | Mat(_) | Bool(_) => Err(InvalidOperand {
+                        expression: expression.clone(),
+                        operand: *function_original.clone(),
+                    }),
+
+                    Fun(_, f) => f(&expression, &arguments_evaluated, context),
+
+                    _ => Ok(FunctionValue(Box::new(function), arguments_evaluated)),
+                }
+            }
+            Integer(_) => Ok(expression),
             Rational(x, _) => Ok(if x.denom().is_one() {
                 Integer(x.numer().clone())
             } else {
-                self.clone()
+                expression
             }),
             Complex(z, representation) => Ok(if z.im.is_zero() {
                 Rational(z.re.clone(), *representation)
             } else {
-                self.clone()
+                expression
             }),
-            Vector(_) => Ok(self.clone()), // TODO: Evaluate each element!
-            Matrix(_) => Ok(self.clone()), // TODO: Evaluate each element!
-            Boolean(_) => Ok(self.clone()),
-            Negation(a) => self.evaluate_step_unary(a, context),
-            Not(a) => self.evaluate_step_unary(a, context),
-            Sum(a, b) => self.evaluate_step_binary(a, b, context),
-            Difference(a, b) => self.evaluate_step_binary(a, b, context),
-            Product(a, b) => self.evaluate_step_binary(a, b, context),
-            Quotient(a, b) => self.evaluate_step_binary(a, b, context),
-            Remainder(a, b) => self.evaluate_step_binary(a, b, context),
-            Power(a, b) => self.evaluate_step_binary(a, b, context),
-            Equal(a, b) => self.evaluate_step_binary(a, b, context),
-            NotEqual(a, b) => self.evaluate_step_binary(a, b, context),
-            LessThan(a, b) => self.evaluate_step_binary(a, b, context),
-            LessThanOrEqual(a, b) => self.evaluate_step_binary(a, b, context),
-            GreaterThan(a, b) => self.evaluate_step_binary(a, b, context),
-            GreaterThanOrEqual(a, b) => self.evaluate_step_binary(a, b, context),
-            And(a, b) => self.evaluate_step_binary(a, b, context),
-            Or(a, b) => self.evaluate_step_binary(a, b, context),
+            Vector(v) => {
+                let mut elements = Vec::new();
+
+                for element in v.iter() {
+                    elements.push(element.evaluate_step(context)?);
+                }
+
+                Ok(Vector(crate::expression::Vector::from_vec(elements)))
+            }
+            VectorElement(vector, i) => {
+                let vector_original = vector;
+                let i_original = i;
+
+                let vector = vector.evaluate_step(context)?;
+                let i = i.evaluate_step(context)?;
+
+                match (vector.typ(), i.typ()) {
+                    (Num(_, _) | Bool(_), _) => Err(InvalidOperand {
+                        expression: expression.clone(),
+                        operand: *vector_original.clone(),
+                    }),
+
+                    (_, Mat(_) | Bool(_)) => Err(InvalidOperand {
+                        expression: expression.clone(),
+                        operand: *i_original.clone(),
+                    }),
+
+                    (Mat(vector), Num(i, _)) => {
+                        if vector.ncols() != 1 {
+                            Err(InvalidOperand {
+                                expression: expression.clone(),
+                                operand: *vector_original.clone(),
+                            })
+                        } else if let Some(i) = i.to_usize() {
+                            if i >= vector.nrows() {
+                                Err(IndexOutOfBounds {
+                                    expression: expression.clone(),
+                                    vector_or_matrix: *vector_original.clone(),
+                                    index: *i_original.clone(),
+                                })
+                            } else {
+                                Ok(vector[(i, 0)].clone())
+                            }
+                        } else {
+                            Err(InvalidOperand {
+                                expression: expression.clone(),
+                                operand: *i_original.clone(),
+                            })
+                        }
+                    }
+
+                    _ => Ok(VectorElement(Box::new(vector), Box::new(i))),
+                }
+            }
+            Matrix(m) => {
+                let mut columns = Vec::new();
+
+                for column in m.column_iter() {
+                    let mut elements = Vec::new();
+
+                    for element in column.iter() {
+                        elements.push(element.evaluate_step(context)?);
+                    }
+
+                    columns.push(crate::expression::Vector::from_vec(elements));
+                }
+
+                Ok(if columns.is_empty() {
+                    Vector(crate::expression::Vector::from_vec(Vec::new()))
+                } else if columns.len() == 1 {
+                    Vector(columns.remove(0))
+                } else {
+                    Matrix(crate::expression::Matrix::from_columns(&columns))
+                })
+            }
+            MatrixElement(matrix, i, j) => {
+                let matrix_original = matrix;
+                let i_original = i;
+                let j_original = j;
+
+                let matrix = matrix.evaluate_step(context)?;
+                let i = i.evaluate_step(context)?;
+                let j = j.evaluate_step(context)?;
+
+                match (matrix.typ(), i.typ(), j.typ()) {
+                    (Num(_, _) | Bool(_), _, _) => Err(InvalidOperand {
+                        expression: expression.clone(),
+                        operand: *matrix_original.clone(),
+                    }),
+
+                    (_, Mat(_) | Bool(_), _) => Err(InvalidOperand {
+                        expression: expression.clone(),
+                        operand: *i_original.clone(),
+                    }),
+
+                    (_, _, Mat(_) | Bool(_)) => Err(InvalidOperand {
+                        expression: expression.clone(),
+                        operand: *j_original.clone(),
+                    }),
+
+                    (Mat(matrix), Num(i, _), Num(j, _)) => {
+                        if let Some(i) = i.to_usize() {
+                            if let Some(j) = j.to_usize() {
+                                if i >= matrix.nrows() {
+                                    Err(IndexOutOfBounds {
+                                        expression: expression.clone(),
+                                        vector_or_matrix: *matrix_original.clone(),
+                                        index: *i_original.clone(),
+                                    })
+                                } else if j >= matrix.ncols() {
+                                    Err(IndexOutOfBounds {
+                                        expression: expression.clone(),
+                                        vector_or_matrix: *matrix_original.clone(),
+                                        index: *j_original.clone(),
+                                    })
+                                } else {
+                                    Ok(matrix[(i, j)].clone())
+                                }
+                            } else {
+                                Err(InvalidOperand {
+                                    expression: expression.clone(),
+                                    operand: *j_original.clone(),
+                                })
+                            }
+                        } else {
+                            Err(InvalidOperand {
+                                expression: expression.clone(),
+                                operand: *i_original.clone(),
+                            })
+                        }
+                    }
+
+                    _ => Ok(MatrixElement(Box::new(matrix), Box::new(i), Box::new(j))),
+                }
+            }
+            Boolean(_) => Ok(expression),
+            Negation(a) => expression.evaluate_step_unary(a, context),
+            Not(a) => expression.evaluate_step_unary(a, context),
+            Sum(a, b) => expression.evaluate_step_binary(a, b, context),
+            Difference(a, b) => expression.evaluate_step_binary(a, b, context),
+            Product(a, b) => expression.evaluate_step_binary(a, b, context),
+            Quotient(a, b) => expression.evaluate_step_binary(a, b, context),
+            Remainder(a, b) => expression.evaluate_step_binary(a, b, context),
+            Power(a, b) => expression.evaluate_step_binary(a, b, context),
+            Equal(a, b) => expression.evaluate_step_binary(a, b, context),
+            NotEqual(a, b) => expression.evaluate_step_binary(a, b, context),
+            LessThan(a, b) => expression.evaluate_step_binary(a, b, context),
+            LessThanOrEqual(a, b) => expression.evaluate_step_binary(a, b, context),
+            GreaterThan(a, b) => expression.evaluate_step_binary(a, b, context),
+            GreaterThanOrEqual(a, b) => expression.evaluate_step_binary(a, b, context),
+            And(a, b) => expression.evaluate_step_binary(a, b, context),
+            Or(a, b) => expression.evaluate_step_binary(a, b, context),
         }
     }
 
@@ -363,6 +594,13 @@ impl Expression {
             Expression::Complex(Complex::i(), RationalRepresentation::Fraction),
         );
 
+        for function in functions() {
+            default_context.insert(
+                function.metadata.name.to_owned(),
+                Expression::Function(function.metadata.name.to_owned(), function.implementation),
+            );
+        }
+
         for (identifier, expression) in context {
             default_context.insert(identifier, expression);
         }
@@ -452,8 +690,46 @@ mod tests {
         t("-2 ^ 4", "-16");
         t("(-2) ^ 4", "16");
         t("0.5 ^ 4", "0.0625");
-        t("987654321123456789 ^ 5", "939777062588963894467852986656442266299580252508947542802086985660852317355013741720482949");
-        t("3 ^ 4 ^ 5", "373391848741020043532959754184866588225409776783734007750636931722079040617265251229993688938803977220468765065431475158108727054592160858581351336982809187314191748594262580938807019951956404285571818041046681288797402925517668012340617298396574731619152386723046235125934896058590588284654793540505936202376547807442730582144527058988756251452817793413352141920744623027518729185432862375737063985485319476416926263819972887006907013899256524297198527698749274196276811060702333710356481");
+        t(
+            "987654321123456789 ^ 5",
+            "939777062588963894467852986656442266299580252508947542802086985660852317355013741720482949",
+        );
+        t(
+            "3 ^ 4 ^ 5",
+            "373391848741020043532959754184866588225409776783734007750636931722079040617265251229993688938803977220468765065431475158108727054592160858581351336982809187314191748594262580938807019951956404285571818041046681288797402925517668012340617298396574731619152386723046235125934896058590588284654793540505936202376547807442730582144527058988756251452817793413352141920744623027518729185432862375737063985485319476416926263819972887006907013899256524297198527698749274196276811060702333710356481",
+        );
+    }
+
+    #[test]
+    fn linear_algebra() {
+        t("[1] + [2]", "[3]");
+        t("[1] - [2]", "[-1]");
+        t("[1] * [2]", "[2]");
+        t("[1] * 2", "[2]");
+        t("1 * [2]", "[2]");
+
+        t("[1, 2] + [3, 4]", "[4, 6]");
+        t("[1, 2] - [3, 4]", "[-2, -2]");
+        t("[1, 2] * [[3, 4]]", "[[3, 4], [6, 8]]");
+        t("[[1, 2]] * [3, 4]", "[11]");
+        t("2 * [3, 4]", "[6, 8]");
+        t("[2, 3] * 4", "[8, 12]");
+
+        t(
+            "[[a, b], [c, d], [e, f]] * [[5, 6], [7, 8]]",
+            "[[a * 5 + b * 7, a * 6 + b * 8], [c * 5 + d * 7, c * 6 + d * 8], [e * 5 + f * 7, e * 6 + f * 8]]",
+        );
+    }
+
+    #[test]
+    fn indices() {
+        t("[a][0]", "a");
+        t("[a, b, c][2]", "c");
+        t("[1 + 2, 2 + 3, 3 + 4, 4 + 5][1 + 2]", "9");
+
+        t("[[a]][0, 0]", "a");
+        t("[[a, b, c], [d, e, f]][1, 2]", "f");
+        t("[[1 + 2, 2 + 3], [3 + 4, 4 + 5]][0 + 0, 0 + 1]", "5");
     }
 
     #[test]
